"""
    Encoding functions for DAG-CBOR codec.
"""

from io import BufferedIOBase, BytesIO
import math
import struct
from typing import Optional, overload, Union

import cid # type: ignore

from .utils import CBOREncodingError, DAGCBOREncodingError, _check_key_compliance

EncodableType = Union[None, bool, int, bytes, str, list, dict, cid.cid.BaseCID, float]

@overload
def encode(data: EncodableType, stream: None = None) -> bytes:
    ... # pragma: no cover

@overload
def encode(data: EncodableType, stream: BufferedIOBase) -> None:
    ... # pragma: no cover

def encode(data: EncodableType, stream: Optional[BufferedIOBase] = None) -> Optional[bytes]:
    """
        Encodes the given `data` with the DAG-CBOR codec.
        If a `stream` is given, the encoded data is written to the stream and nothing is returned.
        Otherwise, the encoded data is written to an internal stream and the bytes are returned at the end.
    """
    if stream is None:
        internal_stream = BytesIO()
        _encode(internal_stream, data)
        return internal_stream.getvalue()
    _encode(stream, data)
    return None

def _encode(stream: BufferedIOBase, value: EncodableType):
    # pylint: disable = too-many-return-statements, too-many-branches
    if isinstance(value, bool): # must go before int check
        # major type 0x7 (additional info 20 and 21)
        _encode_bool(stream, value)
    elif isinstance(value, int):
        # major types 0x0 and 0x1
        _encode_int(stream, value)
    elif isinstance(value, bytes):
        # major type 0x2
        _encode_bytes(stream, value)
    elif isinstance(value, str):
        # major type 0x3
        _encode_str(stream, value)
    elif isinstance(value, list):
        # major type 0x4
        _encode_list(stream, value)
    elif isinstance(value, dict):
        # major type 0x5
        _encode_dict(stream, value)
    elif isinstance(value, cid.cid.BaseCID):
        # major type 0x6
        _encode_cid(stream, value)
    elif value is None:
        # major type 0x7 (additional info 22)
        _encode_none(stream, value)
    elif isinstance(value, float):
        # major type 0x7 (additional info 27)
        _encode_float(stream, value)
    else:
        raise DAGCBOREncodingError(f"Type {type(value)} is not encodable.")

def _encode_head(stream: BufferedIOBase, major_type: int, arg: int):
    if arg < 24:
        # argument value as additional info in leading byte
        head = struct.pack(">B", (major_type<<5)|arg)
    elif arg <= 255:
        # leading byte + 1 byte argument value (additional info = 24)
        head = struct.pack(">BB", (major_type<<5)|24, arg)
    elif arg <= 65535:
        # leading byte + 2 bytes argument value (additional info = 25)
        head = struct.pack(">BH", (major_type<<5)|25, arg)
    elif arg <= 4294967295:
        # leading byte + 4 bytes argument value (additional info = 26)
        head = struct.pack(">BL", (major_type<<5)|26, arg)
    else:
        # leading byte + 8 bytes argument value (additional info = 27)
        head = struct.pack(">BQ", (major_type<<5)|27, arg)
    stream.write(head)

def _encode_int(stream: BufferedIOBase, value: int):
    if value >= 18446744073709551616:
        # unsigned int must be < 2**64
        raise CBOREncodingError("Unsigned integer out of range.")
    if value < -18446744073709551616:
        # negative int must be >= -2**64
        raise CBOREncodingError("Negative integer out of range.")
    if value >= 0:
        # unsigned int
        _encode_head(stream, 0x0, value)
    else:
        # negative int
        _encode_head(stream, 0x1, -1-value)

def _encode_bytes(stream: BufferedIOBase, value: bytes):
    _encode_head(stream, 0x2, len(value))
    stream.write(value)

def _encode_str(stream: BufferedIOBase, value: str):
    try:
        utf8_value: bytes = value.encode("utf-8", errors="strict")
    except UnicodeError as e:
        raise CBOREncodingError("Strings must be valid utf-8 strings.") from e
    _encode_head(stream, 0x3, len(utf8_value))
    stream.write(utf8_value)

def _encode_list(stream: BufferedIOBase, value: list):
    _encode_head(stream, 0x4, len(value))
    for item in value:
        _encode(stream, item)

def _encode_dict(stream: BufferedIOBase, value: dict):
    _check_key_compliance(value)
    # sort keys canonically
    try:
        utf8key_val_pairs = [(k.encode("utf-8", errors="strict"), v)
                             for k, v in value.items()]
    except UnicodeError as e:
        raise CBOREncodingError("Strings must be valid utf-8 strings.") from e
    sorted_utf8key_val_pairs = sorted(utf8key_val_pairs, key=lambda i: i[0])
    # encode key-value pairs (keys already utf-8 encoded)
    _encode_head(stream, 0x5, len(value))
    for utf8k, v in sorted_utf8key_val_pairs:
        _encode_head(stream, 0x3, len(utf8k))
        stream.write(utf8k)
        _encode(stream, v)

def _encode_cid(stream: BufferedIOBase, value: cid.cid.BaseCID):
    _encode_head(stream, 0x6, 42)
    _encode_bytes(stream, value.buffer)

def _encode_bool(stream: BufferedIOBase, value: bool):
    _encode_head(stream, 0x7, 21 if value else 20)

def _encode_none(stream: BufferedIOBase, value: None):
    _encode_head(stream, 0x7, 22)

def _encode_float(stream: BufferedIOBase, value: float):
    if math.isnan(value):
        raise DAGCBOREncodingError("NaN is not an allowed float value.")
    if math.isinf(value):
        if value > 0:
            raise DAGCBOREncodingError("Infinity is not an allowed float value.")
        raise DAGCBOREncodingError("-Infinity is not an allowed float value.")
    # special head, with double encoding for 4B argument value
    head = struct.pack(">Bd", (0x7<<5)|27, value)
    stream.write(head)



"""
    Encodes random items with `dag_cbor.encoding.encode`, encodes them with `cbor2.encoder.dumps` and checks that the two encodings match.
"""
# pylint: disable = global-statement

import cbor2.encoder as cbor2 # type: ignore

from dag_cbor import encode
from dag_cbor.random import rand_list, rand_dict, rand_int, rand_bytes, rand_str, rand_bool_none, rand_float, rand_cid, rand_options

nsamples = 1000

def test_int():
    """
        Encodes random `int` samples with `dag_cbor.encoding.encode`,
        encodes them with `cbor2.encoder.dumps` and checks that the two encodings match.
    """
    global nsamples
    test_data = rand_int(nsamples)
    for i, x in enumerate(test_data):
        error_msg = f"failed at #{i} = {repr(x)}"
        assert cbor2.dumps(x) == encode(x), error_msg

def test_bytes():
    """
        Encodes random `bytes` samples with `dag_cbor.encoding.encode`,
        encodes them with `cbor2.encoder.dumps` and checks that the two encodings match.
    """
    global nsamples
    test_data = rand_bytes(nsamples)
    for i, x in enumerate(test_data):
        error_msg = f"failed at #{i} = {repr(x)}"
        assert cbor2.dumps(x) == encode(x), error_msg

def test_str():
    """
        Encodes random `str` samples with `dag_cbor.encoding.encode`,
        encodes them with `cbor2.encoder.dumps` and checks that the two encodings match.
    """
    global nsamples
    test_data = rand_str(nsamples)
    for i, x in enumerate(test_data):
        error_msg = f"failed at #{i} = {repr(x)}"
        assert cbor2.dumps(x) == encode(x), error_msg

def test_bool_none():
    """
        Encodes random `Optional[bool]` or samples with `dag_cbor.encoding.encode`,
        encodes them with `cbor2.encoder.dumps` and checks that the two encodings match.
    """
    global nsamples
    test_data = rand_bool_none(nsamples)
    for i, x in enumerate(test_data):
        error_msg = f"failed at #{i} = {repr(x)}"
        assert cbor2.dumps(x) == encode(x), error_msg

def test_float():
    """
        Encodes random `float` samples with `dag_cbor.encoding.encode`,
        encodes them with `cbor2.encoder.dumps` and checks that the two encodings match.
    """
    global nsamples
    test_data = rand_float(nsamples)
    for i, x in enumerate(test_data):
        error_msg = f"failed at #{i} = {repr(x)}"
        assert cbor2.dumps(x) == encode(x), error_msg

def test_list():
    """
        Encodes random `list` samples with `dag_cbor.encoding.encode`,
        encodes them with `cbor2.encoder.dumps` and checks that the two encodings match.
    """
    global nsamples
    with rand_options(include_cid=False):
        test_data = rand_list(nsamples)
        for i, x in enumerate(test_data):
            error_msg = f"failed at #{i} = {repr(x)}"
            assert cbor2.dumps(x) == encode(x), error_msg

def test_dict():
    """
        Encodes random `dict` samples with `dag_cbor.encoding.encode`,
        encodes them with `cbor2.encoder.dumps` and checks that the two encodings match.
    """
    global nsamples
    with rand_options(include_cid=False):
        test_data = rand_dict(nsamples)
        for i, x in enumerate(test_data):
            error_msg = f"failed at #{i} = {repr(x)}"
            assert cbor2.dumps(x) == encode(x), error_msg

def test_cid():
    """
        Encodes random CID samples with `dag_cbor.encoding.encode`,
        encodes them with `cbor2.encoder.dumps` and checks that the two encodings match.
    """
    global nsamples
    test_data = rand_cid(nsamples)
    for i, x in enumerate(test_data):
        error_msg = f"failed at #{i} = {repr(x)}"
        assert cbor2.dumps(cbor2.CBORTag(42, x.buffer)) == encode(x), error_msg



"""
    Deconding functions for DAG-CBOR codec.
"""

from io import BufferedIOBase, BytesIO
import math
import struct
from typing import Optional, Tuple, Union

import cid # type: ignore

from .encoding import EncodableType
from .utils import CBORDecodingError, DAGCBORDecodingError

def decode(stream_or_bytes: Union[BufferedIOBase, bytes]) -> EncodableType:
    """
        Decodes and returns a single data item from the given `stream_or_bytes`, with the DAG-CBOR codec.
    """
    if isinstance(stream_or_bytes, bytes):
        stream: BufferedIOBase = BytesIO(stream_or_bytes)
    else:
        stream = stream_or_bytes
    data = _decode_item(stream)
    remaining_bytes = stream.read()
    if len(remaining_bytes) > 0:
        raise DAGCBORDecodingError("Encode and decode must operate on a single top-level CBOR object")
    return data

def _decode_item(stream: BufferedIOBase) -> EncodableType:
    # pylint: disable = too-many-return-statements
    major_type, arg = _decode_head(stream)
    if isinstance(arg, float):
        # float
        assert major_type == 0x7
        if math.isnan(arg):
            raise DAGCBORDecodingError("NaN is not an allowed float value.")
        if math.isinf(arg):
            if arg > 0:
                raise DAGCBORDecodingError("Infinity is not an allowed float value.")
            raise DAGCBORDecodingError("-Infinity is not an allowed float value.")
        return arg
    if major_type == 0x0:
        return arg # unsigned int
    if major_type == 0x1:
        return -1-arg # negative int
    if major_type == 0x2:
        return _decode_bytes(stream, arg)
    if major_type == 0x3:
        return _decode_str(stream, arg)
    if major_type == 0x4:
        return _decode_list(stream, arg)
    if major_type == 0x5:
        return _decode_dict(stream, arg)
    if major_type == 0x6:
        return _decode_cid(stream, arg)
    if major_type == 0x7:
        return _decode_bool_none(stream, arg)
    raise RuntimeError("Major type must be one of 0x0-0x7.")

def _decode_head(stream: BufferedIOBase) -> Tuple[int, Union[int, float]]:
    # read leading byte
    res = stream.read(1)
    if len(res) < 1:
        raise CBORDecodingError("Unexpected EOF while reading leading byte of data item head.")
    leading_byte = res[0]
    major_type = leading_byte >> 5
    additional_info = leading_byte & 0b11111
    # read argument value and return (major_type, arg)
    if additional_info < 24:
        # argument value = additional info
        return (major_type, additional_info)
    if additional_info > 27 or (major_type == 0x7 and additional_info != 27):
        raise DAGCBORDecodingError(f"Invalid additional info {additional_info} in data item head for major type {major_type}.")
    argument_nbytes = 1<<(additional_info-24)
    res = stream.read(argument_nbytes)
    if len(res) < argument_nbytes:
        raise CBORDecodingError(f"Unexpected EOF while reading {argument_nbytes} byte argument of data item head.")
    if additional_info == 24:
        # 1 byte of unsigned int argument value to follow
        return (major_type, res[0])
    if additional_info == 25:
        # 2 bytes of unsigned int argument value to follow
        arg = struct.unpack(">H", res)[0]
        if arg <= 255:
            raise DAGCBORDecodingError(f"Integer {arg} was encoded using 2 bytes, while 1 byte would have been enough.")
        return (major_type, arg)
    if additional_info == 26:
        # 4 bytes of unsigned int argument value to follow
        arg = struct.unpack(">L", res)[0]
        if arg <= 65535:
            raise DAGCBORDecodingError(f"Integer {arg} was encoded using 4 bytes, while 2 bytes would have been enough.")
        return (major_type, arg)
    # necessarily additional_info == 27
    if major_type == 0x7:
        # 8 bytes of float argument value to follow
        return (major_type, struct.unpack(">d", res)[0])
    # 8 bytes of unsigned int argument value to follow
    arg = struct.unpack(">Q", res)[0]
    if arg <= 4294967295:
        raise DAGCBORDecodingError(f"Integer {arg} was encoded using 8 bytes, while 4 bytes would have been enough.")
    return (major_type, arg)

def _decode_bytes(stream: BufferedIOBase, length: int) -> bytes:
    res = stream.read(length)
    if len(res) < length:
        raise CBORDecodingError(f"Unexpected EOF while reading {length} bytes of bytestring.")
    return res

def _decode_str(stream: BufferedIOBase, length: int) -> str:
    res = stream.read(length)
    if len(res) < length:
        raise CBORDecodingError(f"Unexpected EOF while reading {length} bytes of string.")
    return res.decode(encoding="utf-8", errors="strict")

def _decode_list(stream: BufferedIOBase, length: int) -> list:
    l: list = []
    for i in range(length):
        try:
            l.append(_decode_item(stream))
        except CBORDecodingError as e:
            raise CBORDecodingError(f"Error while decoding item #{i} in list of length {length}.") from e
    return l

def _decode_dict(stream: BufferedIOBase, length: int) -> dict:
    d: dict = {}
    for i in range(length):
        try:
            k = _decode_item(stream)
        except CBORDecodingError as e:
            raise CBORDecodingError(f"Error while decoding key #{i} in dict of length {length}.") from e
        if not isinstance(k, str):
            raise DAGCBORDecodingError(f"Key #{i} in dict of length {length} is of type {type(k)}, expected string.")
        try:
            v = _decode_item(stream)
        except CBORDecodingError as e:
            raise CBORDecodingError(f"Error while decoding value #{i} in dict of length {length}.") from e
        d[k] = v
    if len(d) != length:
        raise DAGCBORDecodingError(f"Found only {len(d)} unique keys out of {length} key-value pairs.")
    return d

def _decode_cid(stream: BufferedIOBase, arg: int) -> cid.cid.BaseCID:
    if arg != 42:
        raise DAGCBORDecodingError(f"Error while decoding major type 0x6: tag {arg} is not allowed.")
    try:
        cid_bytes = _decode_item(stream)
    except CBORDecodingError as e:
        raise CBORDecodingError("Error while decoding CID bytes.") from e
    if not isinstance(cid_bytes, bytes):
        raise DAGCBORDecodingError(f"Expected CID bytes, found data of type {type(cid_bytes)} instead.")
    return cid.from_bytes(cid_bytes)

def _decode_bool_none(stream: BufferedIOBase, arg: int) -> Optional[bool]:
    if arg == 20:
        return False
    if arg == 21:
        return True
    if arg == 22:
        return None
    raise DAGCBORDecodingError(f"Error while decoding major type 0x7: simple value {arg} is not allowed.")



"""
    Utility classes and functions.
"""

class CBORError(Exception):
    """
        Parent class for all errors due to the CBOR specification.
    """
    ...

class CBOREncodingError(CBORError):
    """
        Class for encoding errors due to the CBOR specification.
    """
    ...

class CBORDecodingError(CBORError):
    """
        Class for decoding errors due to the CBOR specification.
    """
    ...

class DAGCBORError(CBORError):
    """
        Parent class for all errors due to the DAG-CBOR specification.
    """
    ...

class DAGCBOREncodingError(CBOREncodingError, DAGCBORError):
    """
        Class for encoding errors due to the DAG-CBOR specification.
    """
    ...

class DAGCBORDecodingError(CBORDecodingError, DAGCBORError):
    """
        Class for decoding errors due to the DAG-CBOR specification.
    """
    ...

def _canonical_order_dict(value: dict) -> dict:
    try:
        utf8key_key_val_pairs = [(k.encode("utf-8", errors="strict"), k, v) for k, v in value.items()]
    except UnicodeError as e:
        raise CBOREncodingError("Strings must be valid utf-8 strings.") from e
    sorted_utf8key_key_val_pairs = sorted(utf8key_key_val_pairs, key=lambda i: i[0])
    return {k: v for _, k, v in sorted_utf8key_key_val_pairs}

def _check_key_compliance(value: dict):
    """ Check keys for DAG-CBOR compliance. """
    for k in value.keys():
        if not isinstance(k, str):
            raise DAGCBOREncodingError("Keys for maps must be strings.")
    if len(value.keys()) != len(set(value.keys())):
        raise CBOREncodingError("Keys for maps must be unique.")


def canonical_order_dict(value: dict) -> dict:
    """
        Returns a dictionary with canonically ordered keys,
        according to the DAG-CBOR specification.
    """
    _check_key_compliance(value)
    # sort keys canonically
    return _canonical_order_dict(value)



"""
    Python implementation of the [DAG-CBOR codec specification](https://ipld.io/specs/codecs/dag-cbor/spec/).

    This is a CBOR codec (e.g. see [cbor2](https://github.com/agronholm/cbor2)) which enforces additional
    restrictions and conventions, both in encoding and in decoding:

    - The only tag (major type 6) allowed is the CID tag 42, to be encoded as a two bytes head `0xD82A`
      (`0xD8 == 0b110_11000` means "major type 6 (`0b110 == 6`) with 1 byte of argument (`0b11000 = 24`)",
      while `0x2A` is the number 42).
    - Integers (major types 0 and 1) must be encoded using the minimum possible number of bytes.
    - Lengths for major types 2, 3, 4 and 5 must be encoded in the data item head using the minimum possible number of bytes.
    - Map keys must be strings (major type 3) and must be unique.
    - Map keys must be sorted ascendingly according to the lexicographic ordering of their utf-8 encoded bytes (which is
      the standard ordering of `bytes` objects in Python).
    - Indefinite-length items (bytes, strings, lists or maps) are not allowed.
    - The "break" token is not allowed.
    - The only major type 7 items allowed are 64-bit floats (minor 27) and the simple values `true` (minor 20),
      `false` (minor 21) and `null` (minor 22).
    - The special float values `NaN`, `Infinity` and `-Infinity` are not allowed.
    - Encoding and decoding is only allowed on a single top-level item: back-to-back concatenated items at the top level
      are not allowed.
"""

from .encoding import encode
from .decoding import decode
from . import random



""" setup.py created according to https://packaging.python.org/tutorials/packaging-projects """

import setuptools #type:ignore

with open("README.md", "r") as fh:
    long_description: str = fh.read()

setuptools.setup(
    name="dag-cbor",
    version="0.0.5",
    author="hashberg",
    author_email="sg495@users.noreply.github.com",
    url="https://github.com/hashberg-io/py-dag-cbor",
    description="Python implementation of the DAG-CBOR codec.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(exclude=["test"]),
    classifiers=[ # see https://pypi.org/classifiers/
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.6",
        "Operating System :: OS Independent",
        "Natural Language :: English",
        "Typing :: Typed",
    ],
    package_data={"": [],
                  "dag_cbor": ["dag_cbor/py.typed"],
                 },
    install_requires=[
        "py-cid",
        "py-multihash"
    ],
    include_package_data=True
)
